"""LDAP backend for pre-created accounts.
It"s in the proof-of-concept state.
"""

# vim: foldmethod=indent : tw=100
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=logging-fstring-interpolation, logging-not-lazy, logging-format-interpolation
# pylint: disable=raise-missing-from, missing-docstring, too-few-public-methods

import logging
from enum import Enum, auto
from ldap3 import (
    Server,
    Connection,
    ALL,
    MODIFY_REPLACE,
    MODIFY_DELETE,
    MODIFY_ADD,
    SAFE_SYNC,
    SAFE_RESTARTABLE,
    LDIF,
    AUTO_BIND_NO_TLS,
)

from ldf_adapter.config import CONFIG
from ldf_adapter.results import Failure, Rejection, FatalError
from ldf_adapter.backend import generic

logger = logging.getLogger(__name__)


class Mode(Enum):
    READ_ONLY = auto()
    PRE_CREATED = auto()
    FULL_ACCESS = auto()

    @staticmethod
    def from_str(label):
        label = label.lower()
        if label in ("read_only", "read-only", "readonly"):
            return Mode.READ_ONLY
        elif label in ("pre_created", "pre-created", "precreated"):
            return Mode.PRE_CREATED
        elif label in ("full_access", "full-access", "fullaccess"):
            return Mode.FULL_ACCESS
        else:
            msg = (
                f"Unknown mode '{label}'. Supported modes: "
                f"{[name for name, member in Mode.__members__.items()]}."
            )
            logger.error(msg)
            raise Failure(message=msg)


class LdapSearchResult:
    def __init__(self, ldap_connection, args, kwargs):
        """Perform a search with given arguments and create LdapSearchResult object.

        Initialise fields from the return value of search (connection in SAFE_SYNC mode),
        where the return value is a tuple (status, result, response, request).

        @param ldap_connection: an active ldap3.Connection object
        @param args: list of arguments for ldap3.Connection.search method
        @param kwargs: dictionary of key-value arguments for ldap3.Connection.search method
        """
        logger.debug("=========================")
        logger.debug("Searching in the LDAP:")
        logger.debug(f"base: {args[0]}")
        logger.debug(f"filter: {args[1]}")
        logger.debug(f"kwargs: {kwargs}")
        logger.debug("=========================")
        try:
            ldap_connection.bind()
            search_result = ldap_connection.search(*args, **kwargs)
            if ldap_connection.strategy_type in [SAFE_SYNC, SAFE_RESTARTABLE]:
                self.status = search_result[0]
                self.result = search_result[1]
                self.response = search_result[2]
                self.request = search_result[3]
            else:
                self.status = ldap_connection.response not in [None, []]
                self.result = ldap_connection.result
                self.response = ldap_connection.response
                self.request = ldap_connection.request
        except Exception as e:
            msg = "Error searching in LDAP"
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def found(self):
        return self.status

    def get_attribute(self, attribute_name):
        try:
            value = self.response[0]["attributes"][attribute_name]
            if isinstance(value, list):
                return value[0]
            else:
                return value
        except Exception as e:
            logger.warning(f"Attribute {attribute_name} not found in response: {e}")
            return None

    def get_attributes(self):
        attributes = self.response[0]["attributes"]
        for key, value in attributes.items():
            if isinstance(value, list):
                attributes[key] = value[0]
        return attributes

    def get_attribute_for_all(self, attribute_name):
        try:
            all_entries = []
            for entry in self.response:
                value = entry["attributes"][attribute_name]
                if isinstance(value, list):
                    value = value[0]
                all_entries.append(value)
            return all_entries
        except Exception as e:
            logger.warning(f"Attribute {attribute_name} not found in response: {e}")
            return []


class LdapConnection:
    """Connection to the LDAP server."""

    def __init__(self):
        """Initialise connection to LDAP server."""
        if CONFIG.backend.ldap is None:
            raise FatalError(
                message="LDAP backend is not configured. Please check your configuration."
            )
        self.mode = Mode.from_str(CONFIG.backend.ldap.mode)
        self.user_base = CONFIG.backend.ldap.user_base
        self.group_base = CONFIG.backend.ldap.group_base
        self.attr_oidc_uid = CONFIG.backend.ldap.attribute_oidc_uid
        self.attr_local_uid = CONFIG.backend.ldap.attribute_local_uid
        self.shell = CONFIG.backend.ldap.shell
        self.home_base = CONFIG.backend.ldap.home_base
        self.uid_min = CONFIG.backend.ldap.uid_min
        self.uid_max = CONFIG.backend.ldap.uid_max
        self.gid_min = CONFIG.backend.ldap.gid_min
        self.gid_max = CONFIG.backend.ldap.gid_max

        # initialise connection used to generate LDIFs
        self.ldif_connection = Connection(server=None, client_strategy=LDIF)
        self.ldif_connection.bind()
        self.protocol = "ldap"
        if CONFIG.backend.ldap.tls:
            self.protocol = "ldaps"
        # initialise and bind connection to LDAP server
        try:
            server = Server(
                f"{self.protocol}://{CONFIG.backend.ldap.host}:{CONFIG.backend.ldap.port}",
                get_info=ALL,
            )
            if CONFIG.backend.ldap.admin_user and CONFIG.backend.ldap.admin_password:
                # add SAFE_SYNC, so we get more return values
                self.connection = Connection(
                    server,
                    CONFIG.backend.ldap.admin_user,
                    CONFIG.backend.ldap.admin_password,
                    client_strategy=SAFE_RESTARTABLE,
                    auto_bind=AUTO_BIND_NO_TLS,
                    # collect_usage=True,
                )
            else:
                self.connection = Connection(
                    server,
                    client_strategy=SAFE_RESTARTABLE,
                    auto_bind=AUTO_BIND_NO_TLS,
                    collect_usage=True,
                )
            logger.debug(
                "Connection to LDAP server created: %s", self.connection.result
            )
        except Exception as e:
            msg = f"Could not connect to server {self.protocol}://{CONFIG.backend.ldap.host}:{CONFIG.backend.ldap.port}/"
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def init_nextuidgid(self):
        """Initialise uidNext and gidNext entries in FULL_ACCESS mode
        with values starting in configured range.
        """
        try:
            if self.mode == Mode.FULL_ACCESS:
                search_uid = self.search_next_uid()
                if not search_uid.found():
                    self.connection.bind()
                    result = self.connection.add(
                        f"cn=uidNext,{self.user_base}",
                        object_class=["uidNext"],
                        attributes={"cn": "uidNext", "uidNumber": self.uid_min},
                    )
                    search_uid = self.search_next_uid()
                else:
                    logger.info(
                        f"uidNext already initialised: {search_uid.get_attribute('uidNumber')}."
                    )

                search_gid = self.search_next_gid()
                if not search_gid.found():
                    self.connection.bind()
                    result = self.connection.add(
                        f"cn=gidNext,{self.group_base}",
                        object_class=["gidNext"],
                        attributes={"cn": "gidNext", "gidNumber": self.gid_min},
                    )
                    search_gid = self.search_next_gid()
                else:
                    logger.info(
                        f"gidNext already initialised: {search_gid.get_attribute('gidNumber')}."
                    )
        except Exception as e:
            msg = (
                "Error adding entries in LDAP for tracking available UID and GID values"
            )
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def search_user_by_oidc_uid(self, oidc_uid, attributes=[]):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.user_base}",
                f"(&({self.attr_oidc_uid}={oidc_uid})(objectClass=inetOrgPerson)(objectClass=posixAccount))",
            ],
            {"attributes": attributes},
        )

    def search_user_by_local_username(self, username, get_unique_id=False):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.user_base}",
                f"(&({self.attr_local_uid}={username})(objectClass=inetOrgPerson)(objectClass=posixAccount))",
            ],
            {
                "attributes": (
                    [self.attr_local_uid, self.attr_oidc_uid]
                    if get_unique_id
                    else [self.attr_local_uid]
                )
            },
        )

    def search_group_by_name(self, group_name, attributes=[]):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.group_base}",
                f"(&(cn={group_name})(objectClass=posixGroup))",
            ],
            {"attributes": attributes},
        )

    def search_group_by_gid(self, gid):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.group_base}",
                f"(&(gidNumber={gid})(objectClass=posixGroup))",
            ],
            {"attributes": ["cn"]},
        )

    def search_groups_by_member(self, username):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.group_base}",
                f"(&(memberUid={username})(objectClass=posixGroup))",
            ],
            {"attributes": ["cn", "gidNumber"]},
        )

    def search_next_uid(self):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.user_base}",
                "(&(cn=uidNext)(objectClass=uidNext))",
            ],
            {"attributes": ["uidNumber"]},
        )

    def search_next_gid(self):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.group_base}",
                "(&(cn=gidNext)(objectClass=gidNext))",
            ],
            {"attributes": ["gidNumber"]},
        )

    def is_uid_taken(self, uid):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.user_base}",
                f"(&(uidNumber={uid})(objectClass=posixAccount))",
            ],
            {},
        ).found()

    def is_gid_taken(self, gid):
        return LdapSearchResult(
            self.connection,
            [
                f"{self.group_base}",
                f"(&(gidNumber={gid})(objectClass=posixGroup))",
            ],
            {},
        ).found()

    def get_next_uid(self):
        search_result = self.search_next_uid()
        if search_result.found():
            uid = search_result.get_attribute("uidNumber")
            try:
                uid = int(uid)
            except Exception as e:
                logger.error(f"Could not convert uidNumber to int: {e}")
                raise Failure(message="Could not parse uidNumber from LDAP")

            # make sure uid is not taken already and still in allowed range
            next_uid = uid
            while self.is_uid_taken(next_uid) and next_uid <= self.uid_max:
                next_uid += 1

            # make sure uid still in allowed range
            if next_uid > self.uid_max:
                raise Failure(message="No available UIDs left in configured range.")

            # specify uid in MODIFY_DELETE operation to avoid race conditions
            # the operation will fail if the value has been modified in the meantime
            self.connection.bind()
            result = self.connection.modify(
                f"cn=uidNext,{self.user_base}",
                {"uidNumber": [(MODIFY_DELETE, [uid]), (MODIFY_ADD, [next_uid + 1])]},
            )
            return next_uid

    def get_next_gid(self):
        search_result = self.search_next_gid()
        if search_result.found():
            gid = search_result.get_attribute("gidNumber")
            try:
                gid = int(gid)
            except Exception as e:
                logger.error(f"Could not convert gidNumber to int: {e}")
                raise Failure(message="Could not parse gidNumber from LDAP")

            # make sure gid is not taken already and still in allowed range
            next_gid = gid
            while self.is_gid_taken(next_gid) and next_gid <= self.gid_max:
                next_gid += 1

            # make sure gid still in allowed range
            if next_gid > self.gid_max:
                raise Failure(message="No available GIDs left in configured range.")

            # specify gid in MODIFY_DELETE operation to avoid race conditions
            # the operation will fail if the value has been modified in the meantime
            self.connection.bind()
            result = self.connection.modify(
                f"cn=gidNext,{self.group_base}",
                {"gidNumber": [(MODIFY_DELETE, [gid]), (MODIFY_ADD, [next_gid + 1])]},
            )
            return next_gid

    def add_user(self, userinfo, local_username, primary_group_name):
        """Add an LDAP entry for `local_username` with
        all information from `userinfo`.
        If user exists, a Failure exception is raised.
        """
        try:
            dn = f"uid={local_username},{self.user_base}"
            object_class = ["top", "inetOrgPerson", "posixAccount"]
            attributes = {
                "uid": local_username,
                "uidNumber": self.get_next_uid(),
                "gidNumber": self.search_group_by_name(
                    primary_group_name, attributes=["gidNumber"]
                ).get_attribute("gidNumber"),
                "homeDirectory": f"{self.home_base}/{local_username}",
                "loginShell": self.shell,
                self.attr_local_uid: local_username,
                self.attr_oidc_uid: userinfo.unique_id,
            }
            if userinfo.family_name is not None:
                attributes["sn"] = userinfo.family_name
            if userinfo.given_name is not None:
                attributes["givenName"] = userinfo.given_name
            if userinfo.full_name is not None:
                attributes["cn"] = userinfo.full_name
            if userinfo.email is not None:
                attributes["mail"] = userinfo.email
            self.connection.bind()
            result = self.connection.add(
                dn,
                object_class=object_class,
                attributes=attributes,
            )
            logger.debug(
                f"Added LDAP entry for uid {userinfo.unique_id} with local username {local_username}: {result}"
            )
            return result
        except Exception as e:
            msg = f"Failed to add an LDAP entry for uid {userinfo.unique_id} with local username {local_username}"
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def get_all_user_info(self, unique_id):
        """Return all user info for a given unique_id."""
        search_result = self.search_user_by_oidc_uid(unique_id, attributes=["*"])
        if not search_result.found():
            raise Failure(message=f"User with unique_id {unique_id} not found.")
        return search_result.get_attributes()

    def add_user_ldif(self, userinfo, local_username, primary_group_name):
        """Return LDIF representation for a new user entry for `local_username` with
        all information from `userinfo`.
        If user exists, a Failure exception is raised.
        """
        dn = f"uid={local_username},{self.user_base}"
        object_class = ["top", "inetOrgPerson", "posixAccount"]
        gidNumber = self.search_group_by_name(
            primary_group_name, attributes=["gidNumber"]
        ).get_attribute("gidNumber")
        if not gidNumber:
            gidNumber = ""
        attributes = {
            "uid": local_username,
            "uidNumber": "",
            "gidNumber": gidNumber,
            "homeDirectory": f"{self.home_base}/{local_username}",
            "loginShell": self.shell,
            self.attr_local_uid: local_username,
            self.attr_oidc_uid: userinfo.unique_id,
        }
        if userinfo.family_name is not None:
            attributes["sn"] = userinfo.family_name
        if userinfo.given_name is not None:
            attributes["givenName"] = userinfo.given_name
        if userinfo.full_name is not None:
            attributes["cn"] = userinfo.full_name
        if userinfo.email is not None:
            attributes["mail"] = userinfo.email
        self.ldif_connection.bind()
        self.ldif_connection.add(dn, object_class=object_class, attributes=attributes)
        return self.ldif_connection.response

    def map_user(self, userinfo, local_username):
        """Update the LDAP entry for given `local_username` with
        mapped oidc uid.
        If user doesn't exist, a Failure exception is raised.
        """
        try:
            self.connection.bind()
            return self.connection.modify(
                f"uid={local_username},{self.user_base}",
                {self.attr_oidc_uid: [(MODIFY_REPLACE, [userinfo.unique_id])]},
            )
        except Exception as e:
            msg = f"Failed to modify the LDAP entry for uid {userinfo.unique_id} with local username {local_username}"
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def map_user_ldif(self, userinfo, local_username):
        """Return LDIF representation for updating the LDAP entry for given `local_username` with
        mapped oidc uid. If user doesn't exist, a Failure exception is raised.
        """
        try:
            self.ldif_connection.bind()
            self.ldif_connection.modify(
                f"uid={local_username},{self.user_base}",
                {self.attr_oidc_uid: [(MODIFY_REPLACE, [userinfo.unique_id])]},
            )
            return self.ldif_connection.response
        except Exception as e:
            msg = f"Failed to get LDIF to modify the LDAP entry for uid {userinfo.unique_id} with local username {local_username}"
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def update_user(self, userinfo, local_username):
        """Update the LDAP entry for given `local_username` with
        all information in `userinfo`.
        If user doesn't exist, nothing happens.
        """
        try:
            changes = {
                "homeDirectory": [
                    (MODIFY_REPLACE, [f"{self.home_base}/{local_username}"])
                ],
                "loginShell": [(MODIFY_REPLACE, [self.shell])],
                self.attr_local_uid: [(MODIFY_REPLACE, [local_username])],
                self.attr_oidc_uid: [(MODIFY_REPLACE, [userinfo.unique_id])],
            }
            if userinfo.family_name is not None:
                changes["sn"] = [(MODIFY_REPLACE, [userinfo.family_name])]
            if userinfo.given_name is not None:
                changes["givenName"] = [(MODIFY_REPLACE, [userinfo.given_name])]
            if userinfo.full_name is not None:
                changes["cn"] = [(MODIFY_REPLACE, [userinfo.full_name])]
            if userinfo.email is not None:
                changes["mail"] = [(MODIFY_REPLACE, [userinfo.email])]
            self.connection.bind()
            return self.connection.modify(
                f"uid={local_username},{self.user_base}",
                changes,
            )
        except Exception as e:
            msg = f"Failed to modify the LDAP entry for uid {userinfo.unique_id} with local username {local_username}"
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def delete_user(self, local_username):
        """Delete the LDAP entry for given `local_username`.
        If user doesn't exist, no failure is raised.
        """
        try:
            self.connection.bind()
            return self.connection.delete(f"uid={local_username},{self.user_base}")
        except Exception as e:
            msg = (
                f"Failed to delete the LDAP entry for local username {local_username}."
            )
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def add_user_to_group(self, local_username, group_name):
        """Add a user to group.
        If either of them does not exist, a Failure exception is raised.
        """
        try:
            self.connection.bind()
            self.connection.modify(
                f"cn={group_name},{self.group_base}",
                {
                    "memberUid": [(MODIFY_ADD, [local_username])],
                },
            )
        except Exception as e:
            msg = f"Failed to modify the LDAP entry for group {group_name} with local username {local_username}"
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def add_user_to_group_ldif(self, local_username, group_name):
        """LDIF representation for adding a user to group."""
        self.ldif_connection.bind()
        self.ldif_connection.modify(
            f"cn={group_name},{self.group_base}",
            {
                "memberUid": [(MODIFY_ADD, [local_username])],
            },
        )
        return self.ldif_connection.response

    def remove_user_from_group(self, local_username, group_name):
        """Remove a user from group.
        If either of them does not exist, a Failure exception is raised.
        """
        try:
            self.connection.bind()
            self.connection.modify(
                f"cn={group_name},{self.group_base}",
                {
                    "memberUid": [(MODIFY_DELETE, [local_username])],
                },
            )
        except Exception as e:
            msg = f"Failed to modify the LDAP entry for group {group_name} with local username {local_username}"
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def remove_user_from_group_ldif(self, local_username, group_name):
        """LDIF representation for removing a user from a group."""
        self.ldif_connection.bind()
        self.ldif_connection.modify(
            f"cn={group_name},{self.group_base}",
            {
                "memberUid": [(MODIFY_DELETE, [local_username])],
            },
        )
        return self.ldif_connection.response

    def get_user_groups(self, local_username):
        """Get all groups a user belongs to.

        Returns a list of names.
        """
        logger.debug(f"Searching groups for user {local_username} in LDAP...")
        result = self.search_groups_by_member(local_username)
        if not result.found:
            return []
        return result.get_attribute_for_all("cn")

    def add_group(self, group_name):
        """Add an LDAP entry for `group_name`.
        If group exists, a warning is issued.
        """
        try:
            self.connection.bind()
            return self.connection.add(
                f"cn={group_name},{self.group_base}",
                object_class=["top", "posixGroup"],
                attributes={
                    "cn": group_name,
                    "gidNumber": self.get_next_gid(),
                },
            )
        except Exception as e:
            msg = f"Failed to add an LDAP entry for group {group_name}."
            logger.error(f"{msg}: {e}")
            raise Failure(message=msg)

    def add_group_ldif(self, group_name):
        """LDIF representation for adding an LDAP entry for `group_name`."""
        self.ldif_connection.bind()
        self.ldif_connection.add(
            f"cn={group_name},{self.group_base}",
            object_class=["top", "posixGroup"],
            attributes={
                "cn": group_name,
                "gidNumber": "",
            },
        )
        return self.ldif_connection.response

    def get_all_group_info(self, name):
        """Return all group info for a given name."""
        search_result = self.search_group_by_name(name, attributes=["*"])
        if not search_result.found():
            raise Failure(message=f"Group with name {name} not found.")
        return search_result.get_attributes()

    @staticmethod
    def load():
        """Load config from a file."""
        logger.debug("................... Initialising LDAP connection...")
        ldap = LdapConnection()
        ldap.init_nextuidgid()
        return ldap


LDAP = LdapConnection.load()


class User(generic.User):
    """Manages the user object on the service."""

    def __init__(self, userinfo, **hooks):
        """
        Arguments:
        userinfo -- (type: UserInfo)
        """
        super().__init__(userinfo, **hooks)
        self.userinfo = userinfo
        self.unique_id = userinfo.unique_id
        logger.debug(f"backend processing: {userinfo.unique_id}")
        if self.exists():
            username = self.get_username()
            primary_group = self.get_primary_group()
            logger.debug(
                f"This user does actually exist. The name is: {username} and the primary group is: {primary_group}"
            )
            self.set_username(username)
            self.primary_group = Group(primary_group)
        else:
            self.set_username(userinfo.username)
            self.primary_group = Group(userinfo.primary_group)

        self.ssh_keys = [key["value"] for key in userinfo.ssh_keys]
        self.post_create_script = CONFIG.backend.ldap.post_create_script

    def exists(self):
        """Return whether the user exists on the service.

        If this returns True,  calling `create` should have no effect or raise an error.
        """
        logger.info(f"Check if user exists: {self.unique_id}")
        result = LDAP.search_user_by_oidc_uid(self.unique_id, attributes=[]).found()
        logger.info(f"User {self.unique_id} exists: {result}")
        return result

    def name_taken(self, name):
        """Return whether the username is already taken by another user on the service,
        i.e. if an entry for it exists in the LDAP and it's not mapped to the current oidc uid.

        In 'pre_created' mode, taken means that the username has been mapped to another oidc uid.
        """
        search_result = LDAP.search_user_by_local_username(name, get_unique_id=True)
        if search_result.found():  # there is an entry for name in LDAP
            oidc_uid = search_result.get_attribute(
                LDAP.attr_oidc_uid
            )  # this is the oidc_uid mapped to name
            if (
                LDAP.mode == Mode.PRE_CREATED and oidc_uid is None
            ):  # pre-created but not mapped
                return False
            return oidc_uid != self.unique_id  # name already mapped to another oidc uid
        else:  # no entry for name found in LDAP
            return False

    def get_username(self):
        """Check if a user exists based on unique_id and return the name"""
        return LDAP.search_user_by_oidc_uid(
            self.unique_id, attributes=[LDAP.attr_local_uid]
        ).get_attribute(LDAP.attr_local_uid)

    def set_username(self, username):
        """Set local username on the service."""
        self.name = username

    def get_primary_group(self):
        """Check if a user exists based on unique_id and return the primary group name."""
        gid = LDAP.search_user_by_oidc_uid(
            self.unique_id, attributes=["gidNumber"]
        ).get_attribute("gidNumber")
        return LDAP.search_group_by_gid(gid).get_attribute("cn")

    def get_groups(self):
        """Get a list of names of all service groups that the user belongs to.

        If the user doesn't exist, return an empty list.
        """
        return LDAP.get_user_groups(self.name)

    def get_ldap_entry(self):
        """Get all information about the user stored in LDAP."""
        return LDAP.get_all_user_info(self.unique_id)

    def create(self):
        """Create the user on the service.

        If the user already exists, do nothing or raise an error
        """
        if LDAP.mode == Mode.READ_ONLY:
            msg = (
                f"LDAP backend in read_only mode, new entry cannot be added for user {self.unique_id}."
                f" (local username {self.name})"
            )
            logger.error(msg)
            raise Rejection(
                message=f"{msg} Please contact an administrator to create an account for you."
            )
        elif LDAP.mode == Mode.PRE_CREATED:
            if not LDAP.search_user_by_local_username(
                self.name, get_unique_id=False
            ).found():
                msg = f"Local username {self.name} not found in LDAP for user {self.unique_id}."
                logger.error(msg)
                raise Rejection(
                    message=f"{msg} Please contact an administrator to pre-create this account for you."
                )
            else:
                # we assume that it was checked via name_taken that the username is not already mapped
                LDAP.map_user(self.userinfo, self.name)
        else:  # Mode.FULL_ACCESS
            LDAP.add_user(self.userinfo, self.name, self.primary_group.name)

    def create_tostring(self):
        """Return command (LDIF) for creating user in LDAP.
        If in pre_created mode and a local username exists, only map the user to it.
        """
        if (
            LDAP.mode == Mode.PRE_CREATED
            and LDAP.search_user_by_local_username(
                self.name, get_unique_id=False
            ).found()
        ):
            return LDAP.map_user_ldif(self.userinfo, self.name)
        return LDAP.add_user_ldif(self.userinfo, self.name, self.primary_group.name)

    def update(self):
        """Update all relevant information about the user on the service.

        If the user doesn't exists, behaviour is undefined.
        """
        if LDAP.mode == Mode.READ_ONLY:
            msg = (
                f"LDAP backend in read_only mode, entry for user {self.unique_id} "
                f"cannot be modified."
            )
            logger.warning(msg)
        elif LDAP.mode == Mode.PRE_CREATED:
            msg = (
                f"LDAP backend in pre_created mode, entry for user {self.unique_id} "
                f"cannot be modified."
            )
            logger.warning(msg)
        else:  # Mode.FULL_ACCESS
            LDAP.update_user(self.userinfo, self.name)

    def delete(self):
        """Delete the user on the service.

        If the user doesn"t exists, do nothing or raise an error.
        """
        if LDAP.mode == Mode.READ_ONLY:
            msg = f"LDAP backend in read_only mode, entry for local username {self.name} cannot be deleted."
            logger.error(msg)
            raise Rejection(message=msg)
        elif LDAP.mode == Mode.PRE_CREATED:
            msg = f"LDAP backend in pre_created mode, entry for local username {self.name} cannot be deleted."
            logger.error(msg)
            raise Rejection(message=msg)
        else:  # Mode.FULL_ACCESS
            LDAP.delete_user(self.name)

    def _get_group_lists(self, supplementary_groups=None):
        """Get the names of the groups the user needs to be added to and removed from,
        based on the given supplementary groups and the groups the user is currently in.

        The supplementary_groups MUST also contain the primary group.

        Arguments:
            supplementary_groups (list(Group) | None): list of group names to which the user should belong.

        Returns:
            (list(str), list(str)): list of groups to add, list of groups to remove
        """
        if supplementary_groups is None:
            supplementary_groups_names = []
        else:
            supplementary_groups_names = [group.name for group in supplementary_groups]
        current_groups = self.get_groups()
        groups_to_add = list(set(supplementary_groups_names) - set(current_groups))
        groups_to_remove = list(set(current_groups) - set(supplementary_groups_names))
        return groups_to_add, groups_to_remove

    def mod(self, supplementary_groups=None):
        """Modify the user on the service.

        The state of the user with respect to the provided Arguments after calling this function
        should not depend on the state the user had previously.

        If the user doesn't exists, behaviour is undefined.

        Arguments:
            supplementary_groups (list[Group], optional): the list of groups the user must be part of. Defaults to None.
        """
        groups_to_add, groups_to_remove = self._get_group_lists(supplementary_groups)
        if groups_to_add == [] and groups_to_remove == []:
            logger.debug(
                f"Empty add & remove group lists for user {self.name}. Nothing to do here."
            )
            return [], []

        if LDAP.mode == Mode.READ_ONLY:
            msg = f"LDAP backend in read_only mode, local username {self.name} cannot be added to given groups."
            logger.warning(msg)
            return [], []
        elif LDAP.mode == Mode.PRE_CREATED:
            msg = "LDAP backend in pre_created mode, group {} does not exist so user {} cannot be added to it."
            added_groups = []
            removed_groups = []
            for group in groups_to_add:
                if Group(group).exists():
                    LDAP.add_user_to_group(self.name, group)
                    added_groups.append(group)
                else:
                    logger.warning(msg.format(group, self.name))
            for group in groups_to_remove:
                if Group(group).exists():
                    LDAP.remove_user_from_group(self.name, group)
                    removed_groups.append(group)
                else:
                    logger.warning(msg.format(group, self.name))
            return added_groups, removed_groups
        else:  # Mode.FULL_ACCESS
            for group in groups_to_add:
                LDAP.add_user_to_group(self.name, group)
            for group in groups_to_remove:
                LDAP.remove_user_from_group(self.name, group)
            return groups_to_add, groups_to_remove

    def mod_tostring(self, supplementary_groups=None):
        """LDIF representation for modifying a user to be added and removed from given groups.

        Arguments:
            supplementary_groups (list[Group], optional): the list of groups the user must be part of. Defaults to None.

        Returns:
            str: LDIF containing all user modifications
        """
        groups_to_add, groups_to_remove = self._get_group_lists(supplementary_groups)
        if groups_to_add == [] and groups_to_remove == []:
            logger.debug(
                f"Empty add & remove group lists for user {self.name}. Nothing to do here."
            )
            return ""
        ldifs = []
        for group in groups_to_add:
            ldifs.append(LDAP.add_user_to_group_ldif(self.name, group))
        for group in groups_to_remove:
            ldifs.append(LDAP.remove_user_from_group_ldif(self.name, group))
        return "\n\n".join(ldifs)

    def install_ssh_keys(self):
        """Install users SSH keys on the service.

        No other SSH keys should be active after calling this function.

        If the user doesn't exists, behaviour is undefined.
        """
        # TODO: use ldapPublicKey schema (sshPublicKey attribute) for storing ssh key in LDAP
        pass

    def uninstall_ssh_keys(self):
        """Uninstall the users SSH keys on the service.

        This must uninstall all SSH keys installed with `install_ssh_keys`. It may uninstall SSH
        keys installed by other means.

        If the user doesn't exists, behaviour is undefined.
        """
        # TODO: use ldapPublicKey schema (sshPublicKey attribute) for storing ssh key in LDAP
        pass


class Group(generic.Group):
    """Manages the group object on the service."""

    def __init__(self, name):
        """
        Arguments:
        name -- The name of the group
        """
        self.name = name

    def exists(self):
        """Return whether the group already exists."""
        logger.debug(f"Check if group exists: {self.name}")
        if LDAP.search_group_by_name(self.name).found():
            logger.debug(f"Group {self.name} exists.")
            return True
        else:
            logger.debug(f"Group {self.name} doesn't exist.")
            return False

    def create(self):
        """Create the group on the service.

        If the group already exists, nothing happens.
        """
        if self.exists():
            logger.info(f"Group {self.name} exists.")
        elif LDAP.mode == Mode.READ_ONLY:
            msg = f"LDAP backend in read_only mode, new entry cannot be added for group {self.name}."
            logger.warning(msg)
        elif LDAP.mode == Mode.PRE_CREATED:
            msg = f"LDAP backend in pre_created mode, new entry cannot be added for group {self.name}."
            logger.warning(msg)
        else:  # Mode.FULL_ACCESS
            LDAP.add_group(self.name)

    def create_tostring(self):
        return LDAP.add_group_ldif(self.name)

    def get_ldap_entry(self):
        """Get all information about the group stored in LDAP."""
        return LDAP.get_all_group_info(self.name)



from dataclasses import dataclass
from typing import Optional
import logging
import smtplib
from email.message import EmailMessage
import json
from pathlib import Path
from ldf_adapter.config import CONFIG

from ldf_adapter.notifier import generic
from ldf_adapter.results import FatalError

logger = logging.getLogger(__name__)

NOTIFY_TIMEOUT = 1.2  # seconds


@dataclass
class EmailSettings:
    send_to_template: str
    subject_template: str
    body_template_file: str


# settings for each notification type.
EMAIL_SETTINGS = {
    generic.NotificationType.ADMIN_DEPLOY: EmailSettings(
        send_to_template="${admin_email}",
        subject_template="Request for deployment for user ${unique_id}",
        body_template_file="admin.deploy.template",
    ),
    generic.NotificationType.ADMIN_DEPLOY_UPDATE: EmailSettings(
        send_to_template="${admin_email}",
        subject_template="Updated request for deployment for user ${unique_id}",
        body_template_file="admin.deploy.update.template",
    ),
    generic.NotificationType.ADMIN_UPDATE: EmailSettings(
        send_to_template="${admin_email}",
        subject_template="Request for update for user ${unique_id}",
        body_template_file="admin.update.template",
    ),
    generic.NotificationType.ADMIN_TEST: EmailSettings(
        send_to_template="${admin_email}",
        subject_template="Test email notification on '${hostname}'",
        body_template_file="admin.test.template",
    ),
    generic.NotificationType.USER_DEPLOY: EmailSettings(
        send_to_template="${email}",
        subject_template="Request for deployment to '${hostname}' submitted",
        body_template_file="user.deploy.template",
    ),
    generic.NotificationType.USER_DEPLOY_UPDATE: EmailSettings(
        send_to_template="${email}",
        subject_template="Updated request for deployment to '${hostname}' submitted",
        body_template_file="user.deploy.update.template",
    ),
    generic.NotificationType.USER_UPDATE: EmailSettings(
        send_to_template="${email}",
        subject_template="Request for account update on '${hostname}' submitted",
        body_template_file="user.update.template",
    ),
    generic.NotificationType.UNKNOWN: EmailSettings(
        send_to_template="${admin_email}",
        subject_template="Unknown notification type",
        body_template_file="unknown.template",
    ),
}


class Notifier(generic.Notifier):
    """Implementation of an email notifier for deployment requests."""

    def __init__(self) -> None:
        """Initialise SMTP notifier."""
        if CONFIG.notifier.email is None:
            raise FatalError("Email notifier is not configured")
        self.smtp_server = CONFIG.notifier.email.smtp_server
        self.smtp_port = CONFIG.notifier.email.smtp_port
        self.admin_email = CONFIG.notifier.email.admin_email
        self.sent_from = CONFIG.notifier.email.sent_from
        self.sent_from_password = CONFIG.notifier.email.sent_from_password
        self.use_ssl = CONFIG.notifier.email.use_ssl
        self.templates_dir = CONFIG.notifier.email.templates_dir

    def _build_email(self, send_to: str, subject: str, content: str) -> EmailMessage:
        """Build and email message.

        Args:
            send_to (str): email address to put in "To" field
            subject (str): email subject line
            content (str): content of the email

        Returns:
            EmailMessage: email object
        """
        msg = EmailMessage()
        msg.set_content(content)
        msg["Subject"] = subject
        msg["From"] = self.sent_from
        msg["To"] = send_to
        return msg

    def _send_email(self, email: EmailMessage):
        """Send an email using the configured SMTP settings."""
        try:
            if self.use_ssl:
                server = smtplib.SMTP_SSL(
                    self.smtp_server, self.smtp_port, timeout=NOTIFY_TIMEOUT
                )
            else:
                server = smtplib.SMTP(
                    self.smtp_server, self.smtp_port, timeout=NOTIFY_TIMEOUT
                )
            server.ehlo()
            if self.sent_from_password:
                server.login(self.sent_from, self.sent_from_password)
            server.send_message(email)
            server.close()
            logger.debug(
                "Email with subject '%s' sent successfully to '%s'!",
                email["Subject"],
                email["To"],
            )
        except Exception as ex:
            raise Exception("Could not send email to {%s}: %s", email["To"], ex)

    def notify(
        self,
        notification_type: generic.NotificationType,
        data: dict,
        **_ignored: dict,
    ):
        """Sends out an email notification.

        Args:
            notification_type (generic.NotificationType): type of notification
            data (dict): data to use in email template
        Returns:
            bool: True if notification was sent, False otherwise
        """
        if notification_type == generic.NotificationType.NONE:
            logger.info("Notification type is NONE, not sending email")
            return False
        data = {**data, "admin_email": self.admin_email}
        send_to = generic.NotificationTemplate(
            EMAIL_SETTINGS[notification_type].send_to_template
        ).fill(**data)
        subject = generic.NotificationTemplate(
            EMAIL_SETTINGS[notification_type].subject_template
        ).fill(**data)
        content = generic.NotificationTemplate.load(
            Path(self.templates_dir)
            / EMAIL_SETTINGS[notification_type].body_template_file
        ).fill(**data)
        email = self._build_email(send_to, subject, content)
        self._send_email(email)
        return True

    def test(self):
        """Send test notification to admin to make sure the configured set-up works."""
        if CONFIG.notifier.email is None:
            raise FatalError("Email notifier is not configured")
        settings = CONFIG.notifier.email.to_dict()
        settings["sent_from_password"] = (
            "*****" if settings["sent_from_password"] else None
        )
        data = {
            "admin_email": CONFIG.notifier.email.admin_email,
            "hostname": CONFIG.login_info.ssh_host,
            "notifier": "email",
            "settings": json.dumps(settings, indent=4),
        }
        self.notify(notification_type=generic.NotificationType.ADMIN_TEST, data=data)



"""
Generate useful user or group names
"""

# vim: foldmethod=indent : tw=100
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=logging-fstring-interpolation, logging-not-lazy, logging-format-interpolation
# pylint: disable=raise-missing-from, missing-docstring, too-few-public-methods

import logging
from typing import Optional
from ldf_adapter.config import CONFIG

logger = logging.getLogger(__name__)


class NameGenerator:
    """
    Meta class that returns one of the actual classes, depending on preference
    """

    def __init__(self, generator_mode: str = "friendly", **kwargs):
        if generator_mode == "friendly":
            self.generator = FriendlyNameGenerator(kwargs["userinfo"])
        else:
            self.generator = PooledNameGenerator(kwargs["pool_prefix"])

    def suggest_name(self, *args, **kwargs) -> Optional[str]:
        return self.generator.suggest_name(*args, **kwargs)

    def tried_names(self) -> list:
        if isinstance(
            self.generator,
            FriendlyNameGenerator,
        ):
            return self.generator.tried_names()
        return []


class FriendlyNameGenerator:
    """
    Create Names from UserInfo.
    Don't return the same name twice per run
    """

    strategies = [
        "{self.userinfo.preferred_username:>0}",
        "{self.userinfo.given_name:>0}",
        "{self.userinfo.given_name:.3}{self.userinfo.family_name:.3}",
        "{self.userinfo.family_name:>0}",
        "{self.userinfo.given_name:.4}{self.userinfo.family_name:.3}",
        "{self.userinfo.given_name:.5}{self.userinfo.family_name:.3}",
        "{self.userinfo.given_name:.2}{self.userinfo.family_name:.3}",
        "{self.userinfo.given_name:.4}{self.userinfo.family_name:.4}",
        "{self.userinfo.given_name:.5}{self.userinfo.family_name:.4}",
        "{self.userinfo.given_name:.2}{self.userinfo.family_name:.4}",
        "{self.userinfo.given_name:.4}{self.userinfo.family_name:.5}",
        "{self.userinfo.given_name:.5}{self.userinfo.family_name:.5}",
        "{self.userinfo.given_name:.2}{self.userinfo.family_name:.5}",
        "{self.userinfo.given_name:.4}{self.userinfo.family_name:.2}",
        "{self.userinfo.given_name:.5}{self.userinfo.family_name:.2}",
        "{self.userinfo.given_name:.2}{self.userinfo.family_name:.2}",
        "{self.userinfo.email:>0}",
    ]
    next_strategy_idx = -1

    def __init__(self, userinfo):
        """Generate a useful name"""
        self.userinfo = userinfo
        self.dont_use_these_names = []

    def suggest_name(self, forbidden_names: Optional[list] = None) -> Optional[str]:
        """suggest a valid username"""
        # Copy forbidden names:
        for name in forbidden_names or []:
            if name.lower() not in self.dont_use_these_names:
                self.dont_use_these_names.append(name.lower())

        while True:
            self.next_strategy_idx += 1
            try:
                candidate_name = (
                    self.strategies[self.next_strategy_idx]
                    .format(**locals())
                    .lower()
                    .replace("@", "-")
                )
            except KeyError:
                continue
            except TypeError:
                # e.g. when given_name is None with any formatting (:>0, :.2, etc)
                continue
            except AttributeError as e:
                print(f"ATTRIBUTE ERROR: {e}")
                continue
            except IndexError:
                NL = "\n    "
                logger.error("Ran out of strategies for generating a friendly username")
                logger.error(
                    f"The list of tried usernames is: \n {NL.join(self.dont_use_these_names)}"
                )
                return None

            if candidate_name not in self.dont_use_these_names:
                self.dont_use_these_names.append(candidate_name)
                if CONFIG.messages.log_username_creation:
                    logger.info(f"Potential username: '{candidate_name}'")
                return candidate_name

    def tried_names(self) -> list:
        return self.dont_use_these_names


class PooledNameGenerator:
    """Name Generator for Pooled Accounts"""

    def __init__(self, pool_prefix: str = "pool"):
        self.index = 0
        self.digits = CONFIG.username_generator.pool_digits
        self.username_prefix = CONFIG.username_generator.pool_prefix or pool_prefix
        if self.username_prefix is None:
            self.username_prefix = "pool"

    def suggest_name(self) -> Optional[str]:
        """suggest a valid username"""
        self.index += 1
        if self.index >= 10**self.digits:
            return None
        candidate_name_no_digits = f"{self.username_prefix}"
        potential_length = len(candidate_name_no_digits) + self.digits
        if potential_length > 32:
            excess_chars = potential_length - 32
            candidate_name_no_digits = candidate_name_no_digits[:-excess_chars]
        candidate_name = f"{candidate_name_no_digits}%0{self.digits}d" % self.index
        return candidate_name



# vim: tw=100 foldmethod=expr
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=redefined-outer-name, logging-not-lazy, logging-format-interpolation, logging-fstring-interpolation
# pylint: disable=missing-docstring, trailing-whitespace, trailing-newlines, too-few-public-methods
import pytest
import regex
import random
from pathlib import Path
import subprocess
import os
import logging

import ldf_adapter.backend.local_unix
from ldf_adapter.results import Failure
from ..settings import MockUserInfo

logger = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def local_unix_user(input, exists, taken, monkeypatch):
    """Creates a backend user from provided dict data.
    input should contain:
        - userinfo, which should contain: unique_id, username, primary_group and ssh_keys.
        - new_root: the folder relative to which the user and group dbs will be stored
        - home_base (optional): the base directory for users' home directories
    If exists, it also adds an entry to the user database in /etc/passwd
    Otherwise, if taken, it adds an entry in the user db for this user's username
    """
    # save original subprocess.run for calling inside the mocked one
    old_subprocess_run = subprocess.run

    def mock_root():
        return input["new_root"]

    def mock_subprocess_run(*args, **kwargs):
        """patches calls to system utilities:
        - only for: useradd, userdel, usermod, chage
        - add prefix argument to command
        - patch pkill to do nothing
        - lett all other system calls go through
        """
        logger.debug(args)
        command = args[0]
        if command[0] in ["useradd", "userdel", "usermod", "chage"]:
            new_command = [command[0], "--prefix", mock_root()] + command[1:]
        elif command[0] == "/usr/bin/pkill":
            return None
        else:
            new_command = command
        return old_subprocess_run(new_command, *args[1:], **kwargs)

    with monkeypatch.context() as mp:
        # patch root used by local_unix.User and subprocess.run to use this root for creating users
        mp.setattr("subprocess.run", mock_subprocess_run)
        mp.setattr(
            "ldf_adapter.backend.local_unix.CONFIG.backend.local_unix.shell",
            "/bin/bash",
        )
        mp.setattr("ldf_adapter.backend.local_unix.User.ROOT", mock_root)
        mp.setattr("ldf_adapter.backend.local_unix.Group.ROOT", mock_root)
        if input.get("home_base"):
            mp.setattr(
                "ldf_adapter.backend.local_unix.CONFIG.backend.local_unix.home_base",
                input["home_base"].rstrip("/"),
            )
        # init root and necessary files in new root directory (/etc/{passwd,group,shadow})
        os.makedirs(mock_root())
        os.makedirs(Path(mock_root()) / "etc")
        (Path(mock_root()) / "etc" / "passwd").touch()
        (Path(mock_root()) / "etc" / "group").touch()
        (Path(mock_root()) / "etc" / "shadow").touch()
        if exists:
            (Path(mock_root()) / "etc" / "passwd").write_text(input["passwd_entry"])
        elif taken:
            (Path(mock_root()) / "etc" / "passwd").write_text(input["passwd_taken"])
        (Path(mock_root()) / "etc" / "group").write_text(input["group_entry"])
        # init service user from unix backend
        service_user = ldf_adapter.backend.local_unix.User(
            MockUserInfo(input["userinfo"])
        )
        yield service_user
        # clean up files
        old_subprocess_run(["rm", "-rf", mock_root()])


@pytest.fixture(scope="function")
def local_unix_group(input, exists, monkeypatch):
    """Creates a backend user from provided dict data.
    input should contain:
        - name: the group name (no constraints on allowed names)
        - new_root: the folder relative to which the user and group dbs will be stored
    If exists=True, also adds an entry to the group database in /etc/group
    """
    # save original subprocess.run for calling inside the mocked one
    old_subprocess_run = subprocess.run

    def mock_root():
        return input["new_root"]

    def mock_subprocess_run(*args, **kwargs):
        """patches calls to system utilities:
        - only for: groupadd
        - add prefix argument to command
        - lett all other system calls go through
        """
        logger.debug(args)
        command = args[0]
        if command[0] in ["groupadd"]:
            new_command = [command[0], "--prefix", mock_root()] + command[1:]
        else:
            new_command = command
        return old_subprocess_run(new_command, *args[1:], **kwargs)

    with monkeypatch.context() as mp:
        # patch root used by local_unix.Group and subprocess.run to use this root for creating groups
        mp.setattr("subprocess.run", mock_subprocess_run)
        mp.setattr("ldf_adapter.backend.local_unix.Group.ROOT", mock_root)
        if input.get("home_base"):
            mp.setattr(
                "ldf_adapter.backend.local_unix.CONFIG.backend.local_unix.home_base",
                input["home_base"].rstrip("/"),
            )
        if input.get("punch4nfdi"):
            mp.setattr(
                "ldf_adapter.backend.local_unix.CONFIG.backend.local_unix.punch4nfdi",
                input["punch4nfdi"],
            )
        # init root and necessary files in new root directory (/etc/{passwd,group,shadow})
        os.makedirs(mock_root())
        os.makedirs(Path(mock_root()) / "etc")
        (Path(mock_root()) / "etc" / "group").touch()

        if exists:
            (Path(mock_root()) / "etc" / "group").write_text(input["group_entry"])

        # init service user from unix backend
        service_group = ldf_adapter.backend.local_unix.Group(input["name"])

        yield service_group

        # clean up files
        old_subprocess_run(["rm", "-rf", mock_root()])


INPUT_UNIX = {
    "new_root": f"/tmp/newroot{random.randint(1000, 9999)}",
    "userinfo": {
        "unique_id": "subuid@issuer.domain",
        "primary_group": "testgroup",
        "username": "testuser",
        "ssh_keys": {},
    },
    "passwd_entry": "testuser:x:1000:1000:subuid@issuer.domain:/home/testuser:/bin/bash",
    "passwd_taken": "testuser:x:2000:2000:::",
    "group_entry": "testgroup:x:1000:",
}

INPUT_CUSTOM_HOME_BASE = {
    "new_root": f"/tmp/newroot{random.randint(1000, 9999)}",
    "userinfo": {
        "unique_id": "subuid@issuer.domain",
        "primary_group": "testgroup",
        "username": "testuser",
        "ssh_keys": {},
    },
    "home_base": "/tmp/custom/home/",
    "passwd_entry": "testuser:x:1000:1000:subuid@issuer.domain:/tmp/custom/home/testuser:/bin/bash",
    "group_entry": "testgroup:x:1000:",
}


@pytest.mark.parametrize("input,exists,taken", [(INPUT_UNIX, False, False)])
def test_create(local_unix_user, input):
    """Test that create method adds the appropriate entry in /etc/passwd."""
    local_unix_user.create()
    assert (
        input["passwd_entry"]
        in (Path(input["new_root"]) / "etc" / "passwd").read_text()
    )


@pytest.mark.parametrize("input,exists,taken", [(INPUT_UNIX, False, True)])
def test_create_taken(local_unix_user, input):
    """Test that create raises a Failure if the username is already taken.
    User's primary group needs to exist.
    """
    with pytest.raises(Failure):
        local_unix_user.create()


@pytest.mark.parametrize("input,exists,taken", [(INPUT_CUSTOM_HOME_BASE, False, False)])
def test_create_custom_home_base(local_unix_user, input):
    """Test that create method adds the appropriate entry in /etc/passwd with custom home dir"""
    local_unix_user.create()
    assert (
        input["passwd_entry"]
        in (Path(input["new_root"]) / "etc" / "passwd").read_text()
    )


@pytest.mark.parametrize(
    "input,exists,taken", [(INPUT_UNIX, False, False), (INPUT_UNIX, False, True)]
)
def test_doesnt_exist(local_unix_user):
    """Tests that exists returns False on a clean system, or even if username is taken by another user.
    (i.e. username exists, but the gecos field does not contain this user's unique_id).
    """
    assert not local_unix_user.exists()


@pytest.mark.parametrize("input,exists,taken", [(INPUT_UNIX, True, False)])
def test_exists(local_unix_user, input):
    """Tests that exists returns True if there is an entry in passwd for this unique_id."""
    assert local_unix_user.exists()


@pytest.mark.parametrize("input,exists,taken", [(INPUT_UNIX, False, True)])
def test_name_taken(local_unix_user, input):
    """Tests that name_taken returns True if the username exists on the system but is not
    mapped to this user's unique_id (via gecos field).
    """
    assert local_unix_user.name_taken(input["userinfo"]["username"])


@pytest.mark.parametrize("input,exists,taken", [(INPUT_UNIX, True, False)])
def test_name_taken_byme(local_unix_user, input):
    """Tests that name_taken returns False if the username exists on the system and is
    mapped to this user's unique_id (via gecos field).
    """
    assert not local_unix_user.name_taken(input["userinfo"]["username"])


@pytest.mark.parametrize("input,exists,taken", [(INPUT_UNIX, True, False)])
def test_get_username_same(local_unix_user, input):
    """Tests that get_username returns the username in passwd.
    Case 1: username in userinfo is the same as the one in passwd.
    """
    assert local_unix_user.get_username() == "testuser"


@pytest.mark.parametrize("input,exists,taken", [(INPUT_UNIX, False, False)])
def test_get_username_different(local_unix_user, input):
    """Tests that get_username returns the username in passwd.
    Case 2: username in userinfo is different than the one in passwd.
    """
    passwd_mapped = (
        "testuser2:x:1000:1000:subuid@issuer.domain:/home/testuser:/bin/bash"
    )
    (Path(input["new_root"]) / "etc" / "passwd").write_text(passwd_mapped)
    assert local_unix_user.get_username() == "testuser2"


@pytest.mark.parametrize(
    "input,exists,taken", [(INPUT_UNIX, False, False), (INPUT_UNIX, False, True)]
)
def test_get_username_none(local_unix_user, input):
    """Tests that get_username returns None if there is no username mapped to the
    user's unique_id, even when the username in userinfo is already taken.
    """
    assert local_unix_user.get_username() == None


@pytest.mark.parametrize("input,exists,taken", [(INPUT_UNIX, True, False)])
def test_delete(local_unix_user, input):
    """Test that after delete, there is no user with the username in the db,
    or any other user mapped to the unique_id."""
    local_unix_user.delete()
    passwd_content = (Path(input["new_root"]) / "etc" / "passwd").read_text()
    assert input["userinfo"]["unique_id"] not in passwd_content
    assert input["userinfo"]["username"] not in passwd_content


@pytest.mark.parametrize(
    "input,exists,taken", [(INPUT_UNIX, False, False), (INPUT_UNIX, False, True)]
)
def test_delete_doesnt_exist(local_unix_user, input, taken):
    """Test that delete raises a Failure if the user is not in the db,
    even if its username is taken by another user."""
    with pytest.raises(Failure):
        local_unix_user.delete()
    passwd_content = (Path(input["new_root"]) / "etc" / "passwd").read_text()
    assert input["userinfo"]["unique_id"] not in passwd_content
    if taken:
        assert input["userinfo"]["username"] in passwd_content


INPUT_UNIX_GROUP = {
    "new_root": f"/tmp/newroot{random.randint(1000, 9999)}",
    "name": "testgroup",
    "group_entry": "testgroup:x:1000:",
}


@pytest.mark.parametrize("input,exists", [(INPUT_UNIX_GROUP, False)])
def test_group_create(local_unix_group, input):
    """Test that create method adds the appropriate entry in /etc/group."""
    local_unix_group.create()
    assert (
        input["group_entry"] in (Path(input["new_root"]) / "etc" / "group").read_text()
    )


@pytest.mark.parametrize("input,exists", [(INPUT_UNIX_GROUP, True)])
def test_group_create_exists(local_unix_group):
    """Test that create raises a Failure if the group exists."""
    with pytest.raises(Failure):
        local_unix_group.create()


@pytest.mark.parametrize(
    "input,exists", [(INPUT_UNIX_GROUP, False), (INPUT_UNIX_GROUP, True)]
)
def test_group_exist(local_unix_group, exists):
    """Tests that exists returns True if there is an entry for the given name,
    and False otherwise.
    """
    assert local_unix_group.exists() == exists


INPUT_SHADOW_COMPATIBLE = [
    ("user", "user"),
    ("", "_"),
    ("äöüÄÖÜß!$*@", "aeoeueaeoeuessisx_at_"),
    ("#%^&()=+[]{}\\|;:'\",<.>/?", "________________________"),
    ("u#%^&()=+[]{}\\|;:'\",<.>/?", "u________________________"),
    ("\u5317\u4EB0", "bei_jing_"),
    ("\u20AC", "eur"),
    ("user$", "users"),
    ("-user", "_-user"),
    ("-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", "_-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"),
    ("-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", "_aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"),
    ("-abcdefaaaaaaaaaaaaaaaaaaaaaaaaaa", "_..defaaaaaaaaaaaaaaaaaaaaaaaaaa"),
    ("abcdefaaaaaaaaaaaaaaaaaaaaaaaaaaa", "__defaaaaaaaaaaaaaaaaaaaaaaaaaaa"),
    ("helmholtz-de_KIT_Helmholtz-member", "helmholtz-de_.._helmholtz-member"),
    # ("a_b_c_d_e_f_a_a_a_a______________", "a_.._d_e_f_a_a_a_a______________"), # ??
    # ("_________________________________", "_.._____________________________"), # ??
]

INPUT_SHADOW_COMPATIBLE_PUNCH4NFDI = [
    ("dfn-de-nfdi-de-punch_punch4nfdi_punch", "punch4nfdi_punch"),
    ("dfn-de-nfdi-de-punch_punch4nfdi_punch_intra", "punch4nfdi_punch_intra"),
]

INPUT_SHADOW_COMPATIBLE_FAIL = [
    "dfn-de-nfdi-de-punch_punch4nfdi_punch_intra",
    "dfn-de-nfdi-de-punch_punch4nfdi_punch_intra_physics_biology_chemistry",
    "_________________________________",  # all _ => no fragments can be shortened
]

INPUT_SHADOW_COMPATIBLE_V044 = [
    ("user", "user"),  # valid name
    ("", "_"),  # empty name
    (
        "äöüÄÖÜß!$*@",
        "aeoeueaeoeuessisx_at_",
    ),  # umlauts and few other special characters can be replaced
    (
        "#%^&()=+[]{}\\|;:'\",<.>/?",
        "________________________",
    ),  # all other special chars replaced with _, when first char is special
    (
        "u#%^&()=+[]{}\\|;:'\",<.>/?",
        "u________________________",
    ),  # all other special chars replaced with _, when first char is a letter
    ("\u5317\u4EB0", "bei_jing_"),  # unicode
    ("\u20AC", "eur"),  # unicode
    (
        "user$",
        "users",
    ),  # $ replaced with s even ar the end (although valid shadow name)
    ("-user", "_user"),  # - as first character replaced with _
    (
        "-aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
        "_aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
    ),  # - as first character replaced with _
    # from here on, length > 32 needs to be shortened to 32; fragments are defined as substrings separated by _
    (
        "helmholtz-de_KIT_Helmholtz-member",
        "helmholtz.._kit_helmholtz-member",
    ),  # real world example: all lowercase, first fragment shortened from the end (denoted by ..)
    (
        "-abcdefaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
        "_abcdefaaaaaaaaaaaaaaaaaaaaaaa..",
    ),  # - as first character replaced with _, one fragment, shortened from the end
    (
        "abcdefaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
        "abcdefaaaaaaaaaaaaaaaaaaaaaaaa..",
    ),  # one fragment
    (
        "abcdefghij_abcdefghij_abcdefghij_abcdefghijk",
        "a.._abc.._abcdefghij_abcdefghijk",
    ),  # multiple fragments, two fragments need to be shortened, start with the first fragment; first character is always kept from a fragment
    (
        "abcdefaaaaaaaaaaaaaaaaaaaaaaaaaaa_bbbbbbbbbb",
        "abcdefaaaaaaaaaaaaa.._bbbbbbbbbb",
    ),  # multiple fragments, shortening first fragment is sufficient
    (
        "abcdefghijkl_b_cc__abcdefghijklmnop_eeeeeeee",
        "a.._b_cc__abcdefghijk.._eeeeeeee",
    ),  # fragments with length 0, 1, 2 are not shortened
    (
        "a_ab_abc_abcde_abcde_abcdefghi_abcdefghijklm",
        "a_ab_abc_a.._a.._a.._abcdefghi..",
    ),  # fragments with length 1, 2, 3 are not shortened
    (
        "aaaaaaaaaa_bbbb_cccc_ddddddddddddddddddddddd",
        "a.._b.._c.._dddddddddddddddddd..",
    ),  # all fragments are shortened
    (
        "abcdefabcdefabcdefabcdefabcdefabcdefabcdef_aaaaaaa",
        "abcdefabcdefabcdefabcd.._aaaaaaa",
    ),  # only first fragment is shortened
    (
        "abcdefabcdef_a_ab_cd_ef_abc_abcd_abcdef_abcdefghij",
        "a.._a_ab_cd_ef_abc_a.._a.._abc..",
    ),  # fragments with length 1, 2, 3 are not shortened
]

INPUT_SHADOW_COMPATIBLE_FAIL_V044 = [
    "_________________________________",  # all _ => no fragments can be shortened
    "a_b_c_d_e_f_g_h_i_j_k_l_m_n_o_p_q",  # all fragments of length 1, cannot be shortened
    "aa_bb_cc_dd_ee_ff_gg_hh_ii_jj_kk_",  # all fragments of length 2, cannot be shortened
    "aaa_bbb_ccc_ddd_eee_fff_ggg_hhh_iii",  # all fragments of length 3, cannot be shortened
]


@pytest.mark.parametrize("raw", [x[0] for x in INPUT_SHADOW_COMPATIBLE])
def test_make_shadow_compatible_length(raw):
    """a shadow-compatible name must be at most 32 characters long"""
    assert len(ldf_adapter.backend.local_unix.make_shadow_compatible(raw)) <= 32


@pytest.mark.parametrize(
    "raw", [x[0] for x in INPUT_SHADOW_COMPATIBLE + INPUT_SHADOW_COMPATIBLE_V044]
)
def test_make_shadow_compatible_length_v044(raw):
    assert len(ldf_adapter.backend.local_unix.make_shadow_compatible_v044(raw)) <= 32


@pytest.mark.parametrize("raw", [x[0] for x in INPUT_SHADOW_COMPATIBLE])
def test_make_shadow_compatible_allowed_chars(raw):
    """a shadow-compatible name must start with a lowercase letter or underscore
    and can also contain numbers and - in addition to lowercase letters and _
    """
    word = ldf_adapter.backend.local_unix.make_shadow_compatible(raw)
    assert regex.match(r"[a-z_]", word[0]) and regex.match(r"[-0-9_a-z]", word)


@pytest.mark.parametrize(
    "raw", [x[0] for x in INPUT_SHADOW_COMPATIBLE + INPUT_SHADOW_COMPATIBLE_V044]
)
def test_make_shadow_compatible_allowed_chars_v044(raw):
    word = ldf_adapter.backend.local_unix.make_shadow_compatible_v044(raw)
    assert regex.match(r"[a-z_]", word[0]) and regex.match(r"[-0-9_a-z]", word)


@pytest.mark.parametrize("raw,cooked", INPUT_SHADOW_COMPATIBLE)
def test_make_shadow_compatible(monkeypatch, raw, cooked):
    """expected behaviour:
    - german umlauts are replaced with their phonetic equivalents
    - a few special characters are replaced by sensible equivalents:
        - ! to i
        - $ to s
        - * to x
        - @ to _at_
    - unicode characters are decoded to ascii
    - all other special characters are replaced with _
    - shortening names longer than 32 chars to 32 as follows:
        - fragments (substrings separated by _) are shortened starting with the first fragment
          until the length 32 is reached
        - a fragment is shortened by removing the necessary amount of characters from the end,
          and adding .. at the end to denote the shortening took place, e.g. "abcdef" -> "abc.."
        - the length of any fragment has to be > 3 to be considered for shortening
        - the first character of a fragment is always kept, ie. the strongest shortening of "abcdef" will be "a.."
    """
    monkeypatch.setattr(
        "ldf_adapter.backend.local_unix.CONFIG.backend.local_unix.shadow_compatibility_function",
        "default",
    )
    assert ldf_adapter.backend.local_unix.make_shadow_compatible(raw) == cooked


@pytest.mark.parametrize("raw,cooked", INPUT_SHADOW_COMPATIBLE_V044)
def test_make_shadow_compatible_v044(monkeypatch, raw, cooked):
    monkeypatch.setattr(
        "ldf_adapter.backend.local_unix.CONFIG.backend.local_unix.shadow_compatibility_function",
        "v044",
    )
    assert ldf_adapter.backend.local_unix.make_shadow_compatible(raw) == cooked


@pytest.mark.parametrize("raw,cooked", INPUT_SHADOW_COMPATIBLE_PUNCH4NFDI)
def test_make_shadow_compatible_punch(monkeypatch, raw, cooked):
    monkeypatch.setattr(
        "ldf_adapter.backend.local_unix.CONFIG.backend.local_unix.shadow_compatibility_function",
        "punch",
    )
    assert ldf_adapter.backend.local_unix.make_shadow_compatible(raw) == cooked


@pytest.mark.parametrize("raw,cooked", INPUT_SHADOW_COMPATIBLE_PUNCH4NFDI)
def test_make_shadow_compatible_punch4nfdi(raw, cooked):
    word = ldf_adapter.backend.local_unix.make_shadow_compatible_punch4nfdi(raw)
    assert len(word) <= 32
    assert regex.match(r"[a-z_]", word[0]) and regex.match(r"[-0-9_a-z]", word)
    #  assert word == cooked ## tested above


@pytest.mark.parametrize("raw", INPUT_SHADOW_COMPATIBLE_FAIL)
def test_make_shadow_compatible_fail(raw):
    """expected behaviour: raise ValueError
    - some names might have too many fragments and cannot be shortened
    """
    with pytest.raises(ValueError):
        ldf_adapter.backend.local_unix.make_shadow_compatible(raw)


@pytest.mark.parametrize("raw", INPUT_SHADOW_COMPATIBLE_FAIL_V044)
def test_make_shadow_compatible_fail_v044(raw):
    with pytest.raises(ValueError):
        ldf_adapter.backend.local_unix.make_shadow_compatible_v044(raw)


INPUT_UNIX_GROUP_PUNCH = {
    "new_root": f"/tmp/newroot{random.randint(1000, 9999)}",
    "name": "dfn-de-nfdi-de-punch_punch4nfdi_punch_intra",
    "group_entry": "punch4nfdi_punch_intra:x:1000:",
    "punch4nfdi": True,
}


@pytest.mark.parametrize("input,exists", [(INPUT_UNIX_GROUP_PUNCH, False)])
def test_create_group_punch4nfdi_enabled(local_unix_group, input):
    """Test that create method adds the appropriate entry in /etc/group when punch4nfdi flag is enabled."""
    local_unix_group.create()
    assert (
        input["group_entry"] in (Path(input["new_root"]) / "etc" / "group").read_text()
    )



from typing import Optional
from abc import ABC, abstractmethod

from ldf_adapter.approval.models import PendingUser, PendingGroup, PendingMemberships


class PendingDB(ABC):
    """Generic class to manage users (and groups) in states associated with the approval workflow.

    To add a new implementation for PendingDB, create a new module in this directory containing a
    class also named PendingDB that extends this class and implements all abstract methods.
    The new class will be automatically loaded by the 'databases' factory and can be accessed with
    the name of the module: databases.get("new_db_type")
    """

    @abstractmethod
    def add_user(self, user: PendingUser) -> bool:
        """Add a new entry for a user. Returns False if entry already exists."""
        return False

    @abstractmethod
    def remove_user(self, unique_id: str) -> None:
        """Remove user entry for unique_id."""
        pass

    @abstractmethod
    def get_user(self, unique_id: str) -> Optional[PendingUser]:
        """Get a user entry that is up for approval by the user's unique_id."""
        return None

    @abstractmethod
    def get_user_by_username(self, username: str) -> Optional[PendingUser]:
        """Get a user entry by the local username."""
        return None

    @abstractmethod
    def reject_user(self, unique_id: str) -> None:
        """Change the state of user given by unique_id to REJECTED."""
        pass

    @abstractmethod
    def notify_user(self, unique_id: str) -> None:
        """Change the state of user given by unique_id to NOTIFIED."""
        pass

    @abstractmethod
    def add_group(self, group: PendingGroup) -> bool:
        """Add a new entry for a group. Returns False if entry already exists."""
        return False

    @abstractmethod
    def remove_group(self, name: str) -> None:
        """Remove group entry for name."""
        pass

    @abstractmethod
    def get_group(self, name: str) -> Optional[PendingGroup]:
        """Get a group entry that is up for approval by the group's name."""
        return None

    @abstractmethod
    def notify_group(self, name: str) -> None:
        """Change the state of group given by name to NOTIFIED."""
        pass

    @abstractmethod
    def add_memberships(self, membership: PendingMemberships) -> bool:
        """Add a new entry for a user's pending group memberships. Returns False if entry already exists."""
        return False

    @abstractmethod
    def update_memberships(self, membership: PendingMemberships) -> None:
        """Update a user's pending group memberships by replacing them with the given memberships.
        A user is denoted by the unique_id in membership.
        """
        pass

    @abstractmethod
    def remove_memberships(self, unique_id: str) -> None:
        """Remove all pending group memberships of a given user."""
        pass

    @abstractmethod
    def get_memberships(self, unique_id: str) -> Optional[PendingMemberships]:
        """Get a given user's pending group memberships. Returns None if the user is not found."""
        return None

    @abstractmethod
    def notify_memberships(self, unique_id: str) -> None:
        """Change the state of user's memberships given by unique_id to NOTIFIED."""
        pass



# vim: foldmethod=indent : tw=100
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=wrong-import-order
# pylint: disable=redefined-outer-name, logging-not-lazy, logging-format-interpolation
# pylint: disable=missing-docstring, trailing-whitespace, trailing-newlines, too-few-public-methods

import os
import sys
from dataclasses import dataclass, fields, field
from typing import Optional, List

from feudal_globalconfig import globalconfig
from configparser import ConfigParser
from pathlib import Path
import logging

from ldf_adapter.utils import to_bool, to_int, to_list

logger = logging.getLogger(__name__)


PARSE_CMDLINE_PARAMETERS = True
if "pytest" in sys.modules:
    PARSE_CMDLINE_PARAMETERS = False
else:
    try:
        PARSE_CMDLINE_PARAMETERS = globalconfig.config["parse_commandline_args"]
    except KeyError as e:
        pass

if PARSE_CMDLINE_PARAMETERS:
    from ldf_adapter.cmdline_params import args


def reload_parser() -> ConfigParser:
    """Reload configuration from disk.

    Config locations, by priority:
    --config option (defaults to /etc/feudal/feudal_adapter.conf)
    $FEUDAL_ADAPTER_CONFIG
    ./feudal_adapter.conf
    ~/.config/feudal/feudal_adapter.conf
    /etc/feudal/feudal_adapter.conf

    processing is stopped, once a give file is found
    """

    files = []

    # If the program has arguments with a config_file: prefer it:
    if PARSE_CMDLINE_PARAMETERS:
        files.insert(0, Path(args.config_file))

    # If the caller of the library has provided a configfile: prefer it:
    logger.debug(f"Files: {files}")
    try:
        globalconf_conf_file = Path(globalconfig.config["CONFIGFILE"])
        logger.debug(
            f"Trying config of globalconfig: {globalconfig.config['CONFIGFILE']}"
        )
        if globalconf_conf_file.exists():
            files.insert(0, globalconf_conf_file)
    except KeyError:
        pass

    # Finally, check the environment (last means highes priority)
    filename = os.environ.get("FEUDAL_ADAPTER_CONFIG")
    if filename is None:
        filename = os.environ.get("LDF_ADAPTER_CONFIG")
    if filename:
        files.append(Path(filename))

    # default files
    files += [
        Path("feudal_adapter.conf"),
        Path.home() / ".config" / "feudal_adapter.conf",
        Path.home() / ".config" / "feudal" / "feudal_adapter.conf",
        Path("/etc/feudal/feudal_adapter.conf"),
        Path("ldf_adapter.conf"),
        Path.home() / ".config" / "ldf_adapter.conf",
        Path.home() / ".config" / "feudal" / "ldf_adapter.conf",
        Path("/etc/feudal/ldf_adapter.conf"),
    ]

    cp = ConfigParser()
    config_loaded = False
    for f in files:
        if f.exists():
            files_read = cp.read(f)
            logger.debug(f"Using this config file: {files_read}")
            globalconfig.info = {}
            globalconfig.info["config_files_read"] = files_read
            config_loaded = True
            break
    if not config_loaded:
        logger.warning("Could not find any config file")
        logger.debug("Trying to copy config from globalconfig")
        logger.debug(f"type of CONFIG: {type(cp)}")
        # try:
        #     logger.debug(F"type of globalconfig.config['CONFIG']: {type(globalconfig.config['CONFIG'])}")
        # except KeyError:
        # raise Failure(message="Could not find any config (neither file not in globalconfig)")
        logger.error("Could not find any config (neither file not in globalconfig")
        exit(4)
    return cp


@dataclass
class ConfigSection:
    @classmethod
    def __section__name__(cls):
        return "DEFAULT"

    @classmethod
    def load(cls, config: ConfigParser):
        """Sets only the fields that are present in the config file"""
        try:
            return cls(**config[cls.__section__name__()])
        except KeyError:
            logger.debug(
                "Missing config section %s, using default values.",
                cls.__section__name__(),
            )
            return cls()

    def __post_init__(self):
        """Converts some of the fields to the correct type"""
        for field in fields(self):
            value = getattr(self, field.name)
            if value is None:
                continue
            field_type = field.type
            if field.type.__module__ == "typing":
                if field.type.__str__().startswith(
                    "typing.Optional"
                ) or field.type.__str__().startswith("typing.Union"):
                    field_type = field.type.__args__[0]  # get the type of the field
                elif field.type.__str__().startswith("typing.List"):
                    field_type = list  # treat as a list
                else:
                    return  # no conversion
            # if the field does not have the hinted type, convert it if possible
            if not isinstance(value, field_type):
                if field_type == int:
                    setattr(self, field.name, to_int(value))
                if field_type == bool:
                    setattr(self, field.name, to_bool(value))
                if field_type in [List, List[str], list]:
                    setattr(self, field.name, to_list(value))

    def to_dict(self) -> dict:
        """Converts the config to a dict"""
        return {field.name: getattr(self, field.name) for field in fields(self)}


@dataclass
class ConfigListOfSections:
    @classmethod
    def load(cls, config: ConfigParser):
        """Loads all config sub-sections that start with the given section name"""
        sections = {}
        for field in fields(cls):
            try:
                field_type = field.type.__args__[0]  # assume Optional[ConfigSection]
            except:
                field_type = field.type
            field_name = field.name
            subsection = field_type.__section__name__()
            if subsection in config:
                sections[field_name] = field_type.load(config)
            else:
                logger.debug(f"Missing config section [{subsection}].")
        return cls(**sections)

    def to_dict(self) -> dict:
        """Converts the config to a dict"""
        return {
            field.name: getattr(self, field.name).to_dict()
            for field in fields(self)
            if field is not None
        }


@dataclass
class ConfigLdfAdapter(ConfigSection):
    """Config section for ldf_adapter."""

    backend: str = "local_unix"
    backend_supports_preferring_existing_user: bool = False
    primary_group: Optional[str] = None
    fallback_group: Optional[str] = "nogroup"
    additional_groups: list = field(default_factory=list)
    interactive: bool = False

    @classmethod
    def __section__name__(cls):
        return "ldf_adapter"


@dataclass
class ConfigMessages(ConfigSection):
    """Config section for messages. Selects which information will be logged"""

    log_file: str = "/var/log/feudal/adapter.log"
    log_level: Optional[str] = None
    log_to_console: str = ""
    log_name_changes: bool = True
    log_primary_group_definition: bool = True
    log_username_creation: bool = False

    @classmethod
    def __section__name__(cls):
        return "messages"


@dataclass
class ConfigApproval(ConfigSection):
    """Config section for approval workflow"""

    enabled: bool = False
    user_db_location: str = "/var/lib/feudal/pending_users.db"
    notifier: str = "email"

    @classmethod
    def __section__name__(cls):
        return "approval"


@dataclass
class ConfigEmail(ConfigSection):
    """Config section for email notifier"""

    smtp_server: str = "localhost"
    smtp_port: int = 25
    use_ssl: bool = False
    sent_from: str = "admin@localhost"
    sent_from_password: Optional[str] = None
    admin_email: str = "admin@localhost"
    templates_dir: str = "/etc/feudal/templates"

    @classmethod
    def __section__name__(cls):
        return "notifier.email"


@dataclass
class ConfigCourier(ConfigSection):
    """Config section for courier notifier"""

    api_key: str = "foo"

    @classmethod
    def __section__name__(cls):
        return "notifier.courier"


@dataclass
class ConfigAssurance(ConfigSection):
    """Config section for assurance"""

    prefix: str = "https://refeds.org/assurance/"
    require: str = "profile/cappuccino"
    verified_undeploy: bool = False
    skip: bool = False

    @classmethod
    def __section__name__(cls):
        return "assurance"


@dataclass
class ConfigUsernameGenerator(ConfigSection):
    """Config section for username generation"""

    mode: str = "friendly"
    pool_prefix: Optional[str] = None
    pool_digits: int = 3
    strip_sub_groups: bool = False

    @classmethod
    def __section__name__(cls):
        return "username_generator"


@dataclass
class ConfigLoginInfo(ConfigSection):
    """Config section for login information"""

    description: str = "Local SSH Test Service"
    login_help: str = "Login via `mccli ssh {ssh_host}`"
    ssh_host: str = "localhost"

    @classmethod
    def __section__name__(cls):
        return "login_info"


@dataclass
class ConfigLocalUnix(ConfigSection):
    """Config section for local unix backend"""

    shell: str = "/bin/sh"
    home_base: str = "/home"
    deploy_user_ssh_keys: bool = True
    punch4nfdi: bool = False
    post_create_script: Optional[str] = None
    shadow_compatibility_function: Optional[str] = "default"

    @classmethod
    def __section__name__(cls):
        return "backend.local_unix"

    def __post_init__(self):
        super().__post_init__()
        self.home_base.rstrip("/")


@dataclass
class ConfigBwIdm(ConfigSection):
    """Config section for bwIDM backend"""

    url: str = "https://bwidm-test.scc.kit.edu/rest"
    org_id: str = "fdl"
    http_user: str = "foo"
    http_pass: str = "bar"
    service_name: str = "sshtest"
    log_outgoing_http_requests: bool = False
    post_create_script: Optional[str] = None

    @classmethod
    def __section__name__(cls):
        return "backend.bwidm"


@dataclass
class ConfigLdap(ConfigSection):
    """Config section for ldap backend"""

    mode: str = "read_only"
    host: str = "localhost"
    port: Optional[int] = None
    admin_user: Optional[str] = None
    admin_password: Optional[str] = None
    tls: bool = False
    user_base: str = "ou=users,dc=example"
    group_base: str = "ou=groups,dc=example"
    attribute_oidc_uid: str = "gecos"
    attribute_local_uid: str = "uid"
    shell: str = "/bin/sh"
    home_base: str = "/home"
    post_create_script: Optional[str] = None
    uid_min: int = 1000
    uid_max: int = 60000
    gid_min: int = 1000
    gid_max: int = 60000

    def __post_init__(self):
        super().__post_init__()
        if self.port is None:
            self.port = 636 if self.tls else 1389
        self.home_base.rstrip("/")

    @classmethod
    def __section__name__(cls):
        return "backend.ldap"


@dataclass
class ConfigGroups(ConfigSection):
    """Config section for groups"""

    policy: str = "all"
    method: str = "classic"
    mapping: str = ""
    supported_entitlements: str = ""
    supported_groups: str = ""

    @classmethod
    def __section__name__(cls):
        return "groups"


@dataclass
class ConfigVerboseInfoPlugin(ConfigSection):
    """Config section for verbose plugin"""

    active: bool = False
    filename: str = "/tmp/userinfo/userinfo.json"
    dirname: str = "/tmp/userinfo"

    @classmethod
    def __section__name__(cls):
        return "verbose-info-plugin"


# Add more config sections here by inheriting from ConfigSection and providing a __section__name__
# method that returns the section name.
# Then add the section to either one of the ConfigListOfSections below (if it is a subsection, e.g.
# [backend.<new backend name>]), or directly to the root Configuration class below.


@dataclass
class ConfigNotifiers(ConfigListOfSections):
    """Collection of config sections for all notifiers"""

    email: Optional[ConfigEmail] = None
    courier: Optional[ConfigCourier] = None


@dataclass
class ConfigBackends(ConfigListOfSections):
    """Collection of config sections for all backends"""

    local_unix: ConfigLocalUnix = field(default_factory=ConfigLocalUnix)
    ldap: ConfigLdap = field(default_factory=ConfigLdap)
    bwidm: ConfigBwIdm = field(default_factory=ConfigBwIdm)


@dataclass
class Configuration:
    """All configuration settings for the feudal adapter"""

    ldf_adapter: ConfigLdfAdapter = field(default_factory=ConfigLdfAdapter)
    messages: ConfigMessages = field(default_factory=ConfigMessages)
    approval: ConfigApproval = field(default_factory=ConfigApproval)
    notifier: ConfigNotifiers = field(default_factory=ConfigNotifiers)
    assurance: ConfigAssurance = field(default_factory=ConfigAssurance)
    username_generator: ConfigUsernameGenerator = field(
        default_factory=ConfigUsernameGenerator
    )
    login_info: ConfigLoginInfo = field(default_factory=ConfigLoginInfo)
    backend: ConfigBackends = field(default_factory=ConfigBackends)
    verbose_info_plugin: ConfigVerboseInfoPlugin = field(
        default_factory=ConfigVerboseInfoPlugin
    )
    groups: ConfigGroups = field(default_factory=ConfigGroups)

    @classmethod
    def load(cls, config: ConfigParser):
        """Loads all config settings from the given config parser"""
        return cls(**{f.name: f.type.load(config) for f in fields(cls)})


# Load config on import
CONFIG = Configuration.load(reload_parser())



# vim: tw=100 foldmethod=indent
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=redefined-outer-name, logging-not-lazy, logging-format-interpolation, logging-fstring-interpolation
# pylint: disable=missing-docstring, trailing-whitespace, trailing-newlines, too-few-public-methods

import logging
import os
import sys
from logging.handlers import RotatingFileHandler

logger = logging.getLogger(
    "ldf_adapter"
)  # => This is the key to allow logging from other modules
jsonlogger = logging.getLogger("jsonlog.ldf_adapter")


class PathTruncatingFormatter(logging.Formatter):
    """formatter for logging"""

    def format(self, record):
        pathname = record.pathname
        if len(pathname) > 23:
            pathname = "...{}".format(pathname[-19:])
        record.pathname = pathname
        return super(PathTruncatingFormatter, self).format(record)


def setup_logging():
    """setup logging"""

    # Remove all other logging handlers
    for h in logger.handlers:
        logger.removeHandler(h)

    # Use a nice formatter
    for h in logger.handlers:
        logger.removeHandler(h)

    # formatter = logging.Formatter("[%(asctime)s]%(levelname)8s - %(message)s")
    # if loglevel=="DEBUG":
    formatter = PathTruncatingFormatter(
        "[%(asctime)s] {%(pathname)23s:%(lineno)-3d}%(levelname)8s - %(message)s"
    )

    # set preliminary loglevel from ENV:
    loglevel_env = os.environ.get("LOG")
    stream_handler = logging.StreamHandler()
    if loglevel_env is not None:
        logger.setLevel(loglevel_env)
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

    # Setup logfile
    # pylint: disable=import-outside-toplevel
    from .config import CONFIG

    # Ensure log_file exists and is writeable
    logfile = CONFIG.messages.log_file
    dirname = os.path.dirname(logfile)

    loglevel_env = os.environ.get("LOG", "WARNING")
    loglevel = CONFIG.messages.log_level or loglevel_env
    logger.setLevel(loglevel)

    if not os.path.isdir(os.path.dirname(logfile)) and os.path.dirname(logfile) != "":
        try:
            os.mkdir(dirname, 755)
        except FileExistsError as e:
            # ignore this error
            print(f"Warning file exists: {e}")
        except IOError as e:
            print(f"Error creating dir for logfile: {e}")
            raise
    if not os.path.isfile(logfile):
        try:
            open(logfile, "a").close()
        except IOError as e:
            print(f"Error creating dir for logfile: {e}")
            raise

    # Setup logging to logfile
    file_handler = RotatingFileHandler(logfile, maxBytes=100**6, backupCount=2)
    file_handler.setFormatter(formatter)
    file_handler.setLevel(loglevel)
    logger.addHandler(file_handler)

    # Reconsider logging to console:
    log_to_console = CONFIG.messages.log_to_console
    if log_to_console.lower() in ("false", "no"):
        print(f"removing console logger")
        logger.removeHandler(stream_handler)
    elif log_to_console.lower() in ("true", "yes"):
        print(f"adding console logger")
        logger.setLevel(loglevel)
        stream_handler.setLevel(loglevel_env)
        print(f"console loglevel: {loglevel_env}")
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
    else:
        stream_handler.setLevel(loglevel_env)

    logger.debug(
        f"Running: ----------------------------------------------------------------------------------------------------"
    )
    logger.debug(f'         {" ".join(sys.argv)} ')

    # JSON LOGGER
    # FIXME: jsonlogger name
    jsonlogfile = f"{logfile.rstrip('.log')}-json.log"
    jsonlogger = logging.getLogger("jsondata")
    jsonfile_handler = RotatingFileHandler(jsonlogfile, maxBytes=100**6, backupCount=2)
    jsonfile_handler.setFormatter(formatter)
    jsonfile_handler.setLevel(loglevel)
    jsonlogger.setLevel(loglevel)
    jsonlogger.addHandler(jsonfile_handler)

    return (logger, jsonlogger)


(logger, jsonlogger) = setup_logging()



import pytest
import mock

from ldf_adapter.approval.models import (
    PendingUser,
    PendingGroup,
    PendingMemberships,
    DeploymentState,
)
from ldf_adapter.approval.db import databases

MOCK_USER = PendingUser(
    unique_id="unique_id",
    sub="sub",
    iss="iss",
    email="email@domain",
    full_name="full_name",
    username="username",
    state=DeploymentState.PENDING,
    cmd="cmd",
    infodict={},
)

MOCK_GROUP = PendingGroup(
    name="name",
    state=DeploymentState.PENDING,
    cmd="cmd",
    infodict={},
)

MOCK_MEMBERSHIPS = PendingMemberships(
    unique_id="unique_id",
    supplementary_groups=["group1", "group2"],
    removal_groups=["group3", "group4"],
    state=DeploymentState.PENDING,
    cmd="cmd",
    infodict={},
)

MOCK_MEMBERSHIPS_UPDATED = PendingMemberships(
    unique_id="unique_id",
    supplementary_groups=["group1", "group2", "group3", "group5"],
    removal_groups=["group4"],
    state=DeploymentState.PENDING,
    cmd="cmd",
    infodict={},
)


@pytest.fixture(scope="function")
def pending_db(db_type):
    databases._builders = {}
    if db_type == "sqlite":
        with mock.patch(
            "ldf_adapter.approval.db.sqlite.CONFIG.approval.user_db_location",
            ":memory:",
        ):
            yield databases.get(db_type)
    else:
        yield databases.get(db_type)


@pytest.mark.parametrize("db_type", ["sqlite"])
def test_add_user(pending_db):
    # get_user returns None
    assert pending_db.get_user(MOCK_USER.unique_id) == None
    # successful add
    assert pending_db.add_user(MOCK_USER) == True
    # get_user returns the same user
    assert pending_db.get_user(MOCK_USER.unique_id) == MOCK_USER

    # second add returns False
    assert pending_db.add_user(MOCK_USER) == False


@pytest.mark.parametrize("db_type", ["sqlite"])
def test_remove_user(pending_db):
    # get_user returns None
    assert pending_db.get_user(MOCK_USER.unique_id) == None
    # remove user that doesn't exist
    pending_db.remove_user(MOCK_USER.unique_id)

    # remove user that does exist
    pending_db.add_user(MOCK_USER)
    pending_db.remove_user(MOCK_USER.unique_id)
    # get_user returns None
    assert pending_db.get_user(MOCK_USER.unique_id) == None


@pytest.mark.parametrize("db_type", ["sqlite"])
def test_reject_user(pending_db):
    # get_user returns None
    assert pending_db.get_user(MOCK_USER.unique_id) == None
    # reject user that doesn't exist
    pending_db.reject_user(MOCK_USER.unique_id)

    # reject user that does exist
    pending_db.add_user(MOCK_USER)
    # get_user returns user in PENDING state
    assert pending_db.get_user(MOCK_USER.unique_id).state == DeploymentState.PENDING
    pending_db.reject_user(MOCK_USER.unique_id)
    # get_user returns user in REJECTED state
    assert pending_db.get_user(MOCK_USER.unique_id).state == DeploymentState.REJECTED


@pytest.mark.parametrize("db_type", ["sqlite"])
def test_notify_user(pending_db):
    # get_user returns None
    assert pending_db.get_user(MOCK_USER.unique_id) == None
    # notify user that doesn't exist
    pending_db.notify_user(MOCK_USER.unique_id)

    # notify user that does exist
    pending_db.add_user(MOCK_USER)
    # get_user returns user in PENDING state
    assert pending_db.get_user(MOCK_USER.unique_id).state == DeploymentState.PENDING
    pending_db.notify_user(MOCK_USER.unique_id)
    # get_user returns user in NOTIFIED state
    assert pending_db.get_user(MOCK_USER.unique_id).state == DeploymentState.NOTIFIED


@pytest.mark.parametrize("db_type", ["sqlite"])
def test_add_group(pending_db):
    # get_group returns None
    assert pending_db.get_group(MOCK_GROUP.name) == None
    # successful add
    assert pending_db.add_group(MOCK_GROUP) == True
    # get_group returns the same group
    assert pending_db.get_group(MOCK_GROUP.name) == MOCK_GROUP

    # second add returns False
    assert pending_db.add_group(MOCK_GROUP) == False


@pytest.mark.parametrize("db_type", ["sqlite"])
def test_remove_group(pending_db):
    # get_group returns None
    assert pending_db.get_group(MOCK_GROUP.name) == None
    # remove group that doesn't exist
    pending_db.remove_group(MOCK_GROUP.name)

    # remove group that does exist
    pending_db.add_group(MOCK_GROUP)
    pending_db.remove_group(MOCK_GROUP.name)
    # get_group returns None
    assert pending_db.get_group(MOCK_GROUP.name) == None


@pytest.mark.parametrize("db_type", ["sqlite"])
def test_notify_group(pending_db):
    # get_group returns None
    assert pending_db.get_group(MOCK_GROUP.name) == None
    # notify group that doesn't exist
    pending_db.notify_group(MOCK_GROUP.name)

    # notify group that does exist
    pending_db.add_group(MOCK_GROUP)
    # get_group returns group in PENDING state
    assert pending_db.get_group(MOCK_GROUP.name).state == DeploymentState.PENDING
    pending_db.notify_group(MOCK_GROUP.name)
    # get_group returns group in NOTIFIED state
    assert pending_db.get_group(MOCK_GROUP.name).state == DeploymentState.NOTIFIED


@pytest.mark.parametrize("db_type", ["sqlite"])
def test_add_memberships(pending_db):
    # get_memberships returns None
    assert pending_db.get_memberships(MOCK_MEMBERSHIPS.unique_id) == None
    # successful add
    assert pending_db.add_memberships(MOCK_MEMBERSHIPS) == True
    # get_memberships returns the same memberships
    assert pending_db.get_memberships(MOCK_MEMBERSHIPS.unique_id) == MOCK_MEMBERSHIPS

    # second add returns False
    assert pending_db.add_memberships(MOCK_MEMBERSHIPS) == False


@pytest.mark.parametrize("db_type", ["sqlite"])
def test_remove_memberships(pending_db):
    # get_memberships returns None
    assert pending_db.get_memberships(MOCK_MEMBERSHIPS.unique_id) == None
    # remove memberships that doesn't exist
    pending_db.remove_memberships(MOCK_MEMBERSHIPS.unique_id)

    # remove memberships that does exist
    pending_db.add_memberships(MOCK_MEMBERSHIPS)
    pending_db.remove_memberships(MOCK_MEMBERSHIPS.unique_id)
    # get_memberships returns None
    assert pending_db.get_memberships(MOCK_MEMBERSHIPS.unique_id) == None


@pytest.mark.parametrize("db_type", ["sqlite"])
def test_notify_memberships(pending_db):
    # get_memberships returns None
    assert pending_db.get_memberships(MOCK_MEMBERSHIPS.unique_id) == None
    # notify memberships that don't exist
    pending_db.notify_memberships(MOCK_MEMBERSHIPS.unique_id)

    # notify memberships that do exist
    pending_db.add_memberships(MOCK_MEMBERSHIPS)
    # get_memberships returns memberships in PENDING state
    assert (
        pending_db.get_memberships(MOCK_MEMBERSHIPS.unique_id).state
        == DeploymentState.PENDING
    )
    pending_db.notify_memberships(MOCK_MEMBERSHIPS.unique_id)
    # get memberships returns memberships in NOTIFIED state
    assert (
        pending_db.get_memberships(MOCK_MEMBERSHIPS.unique_id).state
        == DeploymentState.NOTIFIED
    )


@pytest.mark.parametrize("db_type", ["sqlite"])
def test_update_memberships(pending_db):
    # get_memberships returns None
    assert pending_db.get_memberships(MOCK_MEMBERSHIPS.unique_id) == None
    # update memberships that don't exist: doesn't fail, but nothing changes
    pending_db.update_memberships(MOCK_MEMBERSHIPS)
    # get_memberships returns None
    assert pending_db.get_memberships(MOCK_MEMBERSHIPS.unique_id) == None

    # update memberships that do exist
    pending_db.add_memberships(MOCK_MEMBERSHIPS)
    pending_db.update_memberships(MOCK_MEMBERSHIPS_UPDATED)
    # get_memberships returns updated value
    assert (
        pending_db.get_memberships(MOCK_MEMBERSHIPS.unique_id)
        == MOCK_MEMBERSHIPS_UPDATED
    )
